//
//  MTLTimerQueue.m
//  QueueTEst
//
//  Created by narita on 2013/05/23.
//  Copyright (c) 2013年 narita. All rights reserved.
//

#import "MTLTimerQueue.h"

@implementation MTLTimerQueue


#pragma mark -
#pragma mark class methods

+ (id) timerQueueWithTarget:(id)anObject
                     action:(SEL)aSelector
              firstInterval:(NSTimeInterval) firstTime
           continueInterval:(NSTimeInterval)continueTime
{
    return [[[self alloc] initWithTarget:anObject
                                  action:aSelector
                           firstInterval:firstTime
                        continueInterval:continueTime] autorelease];
}

#pragma mark -
#pragma mark init methods

- (id) initWithTarget:(id)anObject
               action:(SEL)aSelector
        firstInterval:(NSTimeInterval)firstTime
     continueInterval:(NSTimeInterval)continueTime
{
    self = [super init];
    
    if (self)
    {
        _queue = [[NSMutableArray alloc] init];
        _timer = nil;
        
        _target = [anObject retain];
        _action = aSelector;
        
        _firstInterval    = firstTime;
        _continueInterval = continueTime;
    }
    
    return self;
}

- (id)init
{    
    return [self initWithTarget:nil
                         action:nil
                  firstInterval:0.0
               continueInterval:0.0];
}

#pragma mark -
#pragma mark dealloc

- (void)dealloc
{
    if( _timer && [_timer isValid] )
    {
        [_timer invalidate];
        _timer = nil;
    }
    
    [_target release];
    _target = nil;
    
    _action = nil;
    
    [super dealloc];
}

#pragma mark -
#pragma mark NSCopying, hash, isEqual:

#pragma mark -
#pragma mark NSCoding methods

#pragma mark -
#pragma mark awakeFromNib-like methods

#pragma mark -
#pragma mark action & canAction methods

- (id)dequeue
{
    id theResult = nil;
    
    @synchronized(_queue)
    {
        if( [_queue count] != 0 )
        {
            theResult = [_queue objectAtIndex:0];
            
            if( theResult )
            {
                [[theResult retain] autorelease];
                [_queue removeObjectAtIndex:0];
            }
        }
    }
    
    return theResult;
}

- (void)enqueue:(id)anObject
{
    if( anObject )
    {
        @synchronized(_queue)
        {
            [_queue addObject:anObject];
            
            // 0から1になったらnotification
            if( [_queue count] == 1 )
            {
                // クライアントがキューにオブジェクトを連続して詰め込む事を考慮して少し時間を空ける
                _timer = [NSTimer scheduledTimerWithTimeInterval:_firstInterval
                                                          target:self
                                                        selector:@selector(fire:)
                                                        userInfo:self
                                                         repeats:NO];
            }
        }
    }
}

- (NSUInteger)count
{
    NSUInteger theResult = 0;
    
    @synchronized(_queue)
    {
        theResult = [_queue count];
    }
    
    return theResult;
}

#pragma mark -
#pragma mark delegate/datasource methods

#pragma mark -
#pragma mark accessor methods (in pairs)

@synthesize target = _target;
@synthesize action = _action;

@synthesize firstInterval    = _firstInterval;
@synthesize continueInterval = _continueInterval;

#pragma mark -
#pragma mark Utility methods

- (void) fire:(NSTimer*)sender
{
    
    _timer = nil;
    
    if( _target && [_target respondsToSelector:_action] )
    {
        id theObject = [self dequeue];
        
        while(theObject)
        {
            [_target performSelector:_action
                          withObject:theObject];
            
            // _continueIntervalがマイナスならばキューに有る物を全て処理する
            if( !(_continueInterval < 0) )
            {
                break;
            }
            
            theObject = [self dequeue];
        }
    }
    
    if( [self count] != 0 )
    {
        // キューに積んであるデータをできるだけ素早く処理する。
        _timer = [NSTimer scheduledTimerWithTimeInterval:_continueInterval
                                                  target:self
                                                selector:@selector(fire:)
                                                userInfo:nil
                                                 repeats:NO];
    }
}

@end
